package com.sbh.background_locations

import android.Manifest
import android.annotation.TargetApi
import android.app.Activity
import android.content.*
import android.content.pm.PackageManager
import android.location.Location
import android.os.Build
import android.os.IBinder
import android.text.format.DateFormat
import android.util.Log
import android.widget.Toast
import androidx.core.app.ActivityCompat
import androidx.localbroadcastmanager.content.LocalBroadcastManager
import com.sbh.background_locations.callbacks.LocationCallback
import com.sbh.background_locations.callbacks.MotionCallback
import com.sbh.background_locations.callbacks.ServiceCallback
import com.sbh.background_locations.models.SBHLocation
import com.sbh.background_locations.utils.SharedPreferenceUtil
import com.sbh.background_locations.utils.SingletonHolder
import java.util.*
import kotlin.collections.HashMap

class BackgroundLocation() {

    private val TAG = "BackgroundLocation"
    private var myReceiver: MyReceiver? = null
    private var mService: LocationUpdatesService? = null
    private var mBound: Boolean = false
    private var mActivity: Activity? = null
    private var mLocationCallback: LocationCallback? = null
    private var mMotionCallback: MotionCallback? = null
    private var mOdometer: Double = 0.0
    private val mRunningQOrLater: Boolean = Build.VERSION.SDK_INT >= Build.VERSION_CODES.Q
    private var mLastLocations: MutableList<SBHLocation> = mutableListOf()
    private var mIsMoving: Boolean = false
    private var mWaitingStart: Calendar? = null
    private var mServiceCallback: ServiceCallback? = null

    init {
        myReceiver = MyReceiver()
    }

    companion object : SingletonHolder<BackgroundLocation, Activity?>(::BackgroundLocation) {
        const val EVENT_LOCATION = "location"
        const val EVENT_MOTION_CHANGE = "motionchange"
        const val REQUEST_PERMISSIONS_REQUEST_CODE = 34
    }

    constructor(activity: Activity?) : this() {
        mActivity = activity
    }

    private val mServiceConnection = object : ServiceConnection {
        override fun onServiceConnected(name: ComponentName, service: IBinder) {
            mBound = true
            val binder = service as LocationUpdatesService.LocalBinder
            mService = binder.service
            Log.d(TAG, "SERVICE CONNECTED")
            mServiceCallback!!.onConnected()
        }

        override fun onServiceDisconnected(name: ComponentName) {
            Log.d(TAG, "SERVICE DISCONNECTED")
            mService = null
            mServiceCallback!!.onDisconnected()
        }
    }

    private fun requestLocation() {
        Log.d(TAG, "REQUESTING LOCATION UPDATES")
        if (!checkPermissions()) {
            requestPermissions()
        } else {
            Log.d(TAG, "requestLocation => Service => ${mService != null}")
            mService!!.startTrackingLocation()
        }
    }

    internal fun onRequestPermissionsResult(requestCode: Int, permissions: Array<out String>?, grantResults: IntArray?): Boolean {
        Log.d(TAG, "onRequestPermissionsResult => Service => ${mService != null}")
        Log.i(TAG, "onRequestPermissionResult")
        if (requestCode == REQUEST_PERMISSIONS_REQUEST_CODE) {
            when {
                grantResults!!.isEmpty() -> Log.i(TAG, "User interaction was cancelled.")
                grantResults[0] == PackageManager.PERMISSION_GRANTED -> mService!!.startTrackingLocation()
                else -> Toast.makeText(mActivity, R.string.permission_denied_explanation, Toast.LENGTH_LONG).show()
            }
        }
        return true

    }

    private fun checkPermissions(): Boolean {
        Log.d(TAG, " checkPermissions() => mRunningQOrLater => $mRunningQOrLater")
        return if (mRunningQOrLater) {
            (PackageManager.PERMISSION_GRANTED == ActivityCompat.checkSelfPermission(mActivity?.applicationContext!!, Manifest.permission.ACCESS_FINE_LOCATION)
                    && PackageManager.PERMISSION_GRANTED == ActivityCompat.checkSelfPermission(mActivity?.applicationContext!!, Manifest.permission.ACCESS_BACKGROUND_LOCATION))
        } else {
            PackageManager.PERMISSION_GRANTED == ActivityCompat.checkSelfPermission(mActivity?.applicationContext!!, Manifest.permission.ACCESS_FINE_LOCATION)
        }
    }

    @TargetApi(Build.VERSION_CODES.P)
    private fun requestPermissions() {
        val shouldProvideRationale = mActivity?.let {
            ActivityCompat.shouldShowRequestPermissionRationale(it, Manifest.permission.ACCESS_FINE_LOCATION)
                    || (mRunningQOrLater && ActivityCompat.shouldShowRequestPermissionRationale(it, Manifest.permission.ACCESS_BACKGROUND_LOCATION))
        }
        if (shouldProvideRationale!!) {

            Log.i(TAG, "Displaying permission rationale to provide additional context.")
            Toast.makeText(mActivity, R.string.permission_rationale, Toast.LENGTH_LONG).show()

        } else {
            Log.i(TAG, "Requesting permission")
            if (mRunningQOrLater) {
                mActivity?.let {
                    ActivityCompat.requestPermissions(it,
                            arrayOf(Manifest.permission.ACCESS_FINE_LOCATION, Manifest.permission.ACCESS_BACKGROUND_LOCATION),
                            REQUEST_PERMISSIONS_REQUEST_CODE)
                }
            } else {
                mActivity?.let {
                    ActivityCompat.requestPermissions(it,
                            arrayOf(Manifest.permission.ACCESS_FINE_LOCATION),
                            REQUEST_PERMISSIONS_REQUEST_CODE)
                }
            }
        }
    }

    private fun bindService() {
        Log.d(TAG, "bindService() => mBound => $mBound")
        mActivity?.let {
            LocalBroadcastManager.getInstance(it.applicationContext).registerReceiver(myReceiver!!,
                    IntentFilter(LocationUpdatesService.ACTION_FOREGROUND_ONLY_LOCATION_BROADCAST))
            if (!mBound) {
                Log.d(TAG, "SERVICE BOUNDED")
                it.bindService(Intent(it, LocationUpdatesService::class.java), mServiceConnection, Context.BIND_AUTO_CREATE)
            }
        }
    }

    fun start() {
        Log.d(TAG, "START => ${mActivity == null}")

        if (mActivity?.applicationContext?.let { SharedPreferenceUtil.getLocationTrackingPref(it) }!!) {
            requestPermissions()
        }
        if (mService == null) {
            mServiceCallback = object : ServiceCallback {
                override fun onConnected() {
                    requestLocation()
                    mServiceCallback = null
                }

                override fun onDisconnected() {
                    mServiceCallback = null
                }

            }
            bindService()
        } else
            requestLocation()
    }

    fun stop() {
        mService!!.stopTrackingLocation()

        mActivity?.applicationContext?.let { LocalBroadcastManager.getInstance(it).unregisterReceiver(myReceiver!!) }

        if (mBound) {
            try {
                mActivity?.unbindService(mServiceConnection)
            } catch (e: Exception) {
                Log.e(TAG, "stop() => ${e.message}")
            }
            mBound = false
        }
    }

    fun onLocation(locationCallback: LocationCallback) {
        mLocationCallback = locationCallback
    }

    fun onMotionChange(motionCallback: MotionCallback) {
        mMotionCallback = motionCallback
    }

    fun onActivityDestroy() {
//        mLocationCallback = null
    }

    fun getOdometer(): Double {
        return mOdometer
    }

    fun setOdometer(odometer: Double, callback: LocationCallback) {
        mOdometer = odometer
        if (mLastLocations.isNotEmpty())
            callback.onLocation(mLastLocations.last())
        else
            callback.onError(0)
    }

    fun getCurrentPosition(callback: LocationCallback) {
        Log.d(TAG, "getCurrentPosition")

        if (mService == null) {
            mServiceCallback = object : ServiceCallback {
                override fun onConnected() {
                    mService?.getCurrentLocation(object : LocationCallback {
                        override fun onLocation(location: SBHLocation) {
                            callback.onLocation(location)
                        }

                        override fun onError(error: Int) {
                            callback.onError(error)
                        }
                    })
                    mServiceCallback = null
                }

                override fun onDisconnected() {
                    mServiceCallback = null
                }

            }
            bindService()
        } else {
            mService?.getCurrentLocation(object : LocationCallback {
                override fun onLocation(location: SBHLocation) {
                    callback.onLocation(location)
                }

                override fun onError(error: Int) {
                    callback.onError(error)
                }
            })
        }
    }

    private fun incrementOdometer(distance: Float) {
        mOdometer = mOdometer.plus(distance)
    }

    private inner class MyReceiver : BroadcastReceiver() {
        private var lastLocation: Location? = null
        override fun onReceive(context: Context, intent: Intent) {
            val location = intent.getParcelableExtra<Location>(LocationUpdatesService.EXTRA_LOCATION)
            if (location != null) {
                val locationMap = HashMap<String, Any>()
                locationMap["timestamp"] = DateFormat.format("yyyy-MM-dd HH:mm:ss", location.time)

                if (lastLocation != null) {
                    val distance = lastLocation!!.distanceTo(location)
                    locationMap["odometer"] = distance
                    locationMap["isMoving"] = distance > 3

                    if (locationMap["isMoving"] as Boolean)
                        incrementOdometer(distance)
                }

                lastLocation = location
                val sbhLocation = SBHLocation(location = location, locationMap = locationMap)
                mLocationCallback?.onLocation(sbhLocation)

                addLocation(sbhLocation)
                checkMovement()
            }
        }

        private fun addLocation(sbhLocation: SBHLocation) {
            if (mLastLocations.size >= 10) {
                mLastLocations.removeAt(0)
            }

            mLastLocations.add(sbhLocation)
            Log.d(TAG, "addLocation() => ${mLastLocations.size}")
        }

        private fun checkMovement() {
            mLastLocations.last().isMoving?.let { mMotionCallback?.onMotionChange(it) }
//            TODO implement => distance, duration, accuracy
//            for (location in mLastLocations) {
//
//            }
//
//
//            val items = mLastLocations.filter {
//                it.isMoving != null && !it.isMoving!!
//            }
//
//            val notMoving = items.size >= 7
//
//            Log.d(TAG, "items => ${items.size}")
//            Log.d(TAG, "notMoving => $notMoving")
//            Log.d(TAG, "mIsMoving => $mIsMoving")
//
//            if(items.size >= 2 && mWaitingStart == null) {
//                Log.d(TAG, "TIMER STARTED")
//                mWaitingStart = Calendar.getInstance()
//            } else if(items.size < 3 && mWaitingStart != null) {
//                Log.d(TAG, "TIMER STOPPED => START MOVING")
//                mWaitingStart = null
//            }
//
//            if(items.size >= 8 && mIsMoving != notMoving) {
//                mIsMoving = notMoving
//                mMotionCallback?.onMotionChange(mIsMoving)
//            } else {
//                mIsMoving = notMoving
//                mMotionCallback?.onMotionChange(mIsMoving)
//            }
        }
    }
}